from unittest.mock import MagicMock

import ops
import pytest
import scenario

from litmus_libs.status_manager import StatusManager


@pytest.mark.parametrize(
    "spec, expected_missing",
    (
        ({"foo": None, "bar": None, "baz": None}, "foo, bar, baz"),
        ({"foo": None, "bar": None, "baz": 42}, "foo, bar"),
        ({"foo": True, "bar": None, "baz": 42}, "bar"),
        ({"foo": True, "bar": [42], "baz": 42}, ""),
    )

)
def test_manager_sets_waiting_on_config_missing(spec, expected_missing):
    sm = StatusManager(MagicMock(), wait_for_config=spec)
    event = MagicMock()
    sm.collect_status(event)
    if not expected_missing:
        assert not event.add_status.called
    else:
        assert event.add_status.called
        call_args = event.add_status.call_args
        expected_msg = sm._configs_missing_msg_template.format(expected_missing)
        assert call_args.args[0].message == expected_msg


@pytest.mark.parametrize(
    "required, present, expected_missing",
    (
        (["foo", "bar"], ["foo"], "bar"),
        (["foo", "bar", "baz"], ["foo"], "bar, baz"),
        (["foo", "bar", "baz"], ["foo", "bar", "baz"], ""),
    )

)
def test_manager_sets_waiting_on_relations_missing(required, present, expected_missing):
    charm = MagicMock()
    charm.model.get_relation = lambda name: name in present
    sm = StatusManager(charm, block_if_relations_missing=required)
    event = MagicMock()
    sm.collect_status(event)
    if not expected_missing:
        assert not event.add_status.called
    else:
        assert event.add_status.called
        call_args = event.add_status.call_args
        expected_msg = sm._relations_missing_msg_template.format(expected_missing)
        assert call_args.args[0].message == expected_msg



@pytest.mark.parametrize(
    "required, passing, expected_failing",
    (
        (["foo", "bar"], ["foo"], "bar"),
        (["foo", "bar", "baz"], ["foo"], "bar, baz"),
        (["foo", "bar", "baz"], ["foo", "bar", "baz"], ""),
    )

)
def test_manager_sets_blocked_on_checks_failing(required, passing, expected_failing):
    charm = MagicMock()
    check_status = {name: ops.pebble.CheckInfo(name,
                                               level=ops.pebble.CheckLevel.READY,
                                               status=(ops.pebble.CheckStatus.UP if name in passing else ops.pebble.CheckStatus.DOWN))
                    for name in required}
    charm.unit.get_container.return_value.get_checks = lambda *_: check_status
    sm = StatusManager(charm, block_if_pebble_checks_failing={"container": required})
    event = MagicMock()
    sm.collect_status(event)
    if not expected_failing:
        assert not event.add_status.called
    else:
        assert event.add_status.called
        call_args = event.add_status.call_args
        expected_msg = sm._checks_failing_msg_template.format(expected_failing)
        assert call_args.args[0].message == expected_msg



@pytest.fixture
def ctx():
    class MyCharm(ops.CharmBase):
        META = {"name": "bartlomiej",
                "containers": {"container1": {}},
                "requires": {"rel1": {"interface": "bar"}},
                }
        CONFIG={"options":{"cfg1": {"type": "boolean"}}}

        def __init__(self, f):
            super().__init__(f)
            self.framework.observe(self.on.collect_unit_status, self._collect_status)

        def _collect_status(self, e):
            StatusManager(self,
                          wait_for_config={"cfg1": self.config.get('cfg1'), "bar": 1, "baz": 0},
                          block_if_pebble_checks_failing={"container1": ("check1", )},
                          block_if_relations_missing=['rel1']
                          ).collect_status(e)
            e.add_status(ops.ActiveStatus("happy status!"))

    ctx = scenario.Context(MyCharm, meta=MyCharm.META, config=MyCharm.CONFIG)
    return ctx


@pytest.mark.parametrize("fail", ("checks", "relation", "config", None))
def test_status_mgr_charm_api_checks(ctx, fail):
    check_status = ops.pebble.CheckStatus.DOWN if fail == "checks" else ops.pebble.CheckStatus.UP

    state_in = scenario.State(
        config={} if fail =="config" else {"cfg1": True},
        relations= {} if fail == "relation" else {scenario.Relation("rel1")},
        containers={
            scenario.Container("container1",
                               _base_plan={"checks": {"check1":{"threshold": 3,
                                                                "startup": "enabled",
                                                                "level": None}}},
                               check_infos={scenario.CheckInfo("check1", status=check_status)})
        }
    )
    state_out = ctx.run(ctx.on.update_status(), state=state_in)

    match fail:
        case "checks":
            assert state_out.unit_status.message == StatusManager._checks_failing_msg_template.format("check1")
        case "relation":
            assert state_out.unit_status.message == StatusManager._relations_missing_msg_template.format("rel1")
        case "config":
            assert state_out.unit_status.message == StatusManager._configs_missing_msg_template.format("cfg1")
        case None:
            assert state_out.unit_status.message == "happy status!"




# Copyright 2025 Canonical Ltd.
# See LICENSE file for licensing details.
"""Generic self-monitoring integration wrappers for all litmus charms."""

from typing import Dict, Optional

import ops
import ops_tracing
from charms.loki_k8s.v1.loki_push_api import LogForwarder
from charms.tempo_coordinator_k8s.v0.tracing import TracingEndpointRequirer

_DEFAULT_ENDPOINT_MAPPING = {
    "charm-tracing": "charm-tracing",
    "logging": "logging",
}


class SelfMonitoring:
    """Self-monitoring relation integrator for all litmus charms.

    Automatically adds charm-tracing and pebble log forwarding (for all workload containers)
    integrations

    Usage:
    >>>    class MyCharm(CharmBase):
    >>>        def __init__(self, *a, **kw):
    >>>            super().__init__(*a, **kw)
    >>>            self._self_monitoring = SelfMonitoring(
    >>>                 self,
    >>>                 endpoint_overrides={
    >>>                     "charm-tracing": "send-charm-traces",
    >>>                 },
    >>>             )
    """

    _expected_interfaces = {
        "logging": "loki_push_api",
        "charm-tracing": "tracing",
    }

    def __init__(
        self,
        charm: ops.CharmBase,
        endpoint_overrides: Optional[Dict[str, str]] = None,
    ):
        endpoint_mapping = _DEFAULT_ENDPOINT_MAPPING.copy()
        if endpoint_overrides:
            endpoint_mapping.update(endpoint_overrides)

        self._validate_endpoints(charm, endpoint_mapping)

        # this injects a pebble-forwarding layer in all sidecars that this charm owns
        self._log_forwarder = LogForwarder(charm, relation_name=endpoint_mapping["logging"])
        # this sets up charm tracing with ops.tracing.
        self._charm_tracing = TracingEndpointRequirer(
            charm,
            relation_name=endpoint_mapping["charm-tracing"],
            protocols=["otlp_http"],
        )

    def _validate_endpoints(self, charm: ops.CharmBase, endpoint_mapping: Dict[str, str]):
        # verify that the charm's metadata has declared all endpoints we're being given here
        for internal_name, custom_name in endpoint_mapping.items():
            ep_meta = charm.meta.requires.get(custom_name, charm.meta.provides.get(custom_name))
            if ep_meta is None:
                # probably you passed an endpoint name to endpoint_mapping,
                # but forgot to add it to charmcraft.yaml or misspelled it
                raise ValueError(
                    f"Charm metadata is missing a required endpoint: {custom_name}({internal_name})"
                )

            if ep_meta.interface_name != self._expected_interfaces[internal_name]:
                raise ValueError(
                    f"Declared charm endpoint {custom_name}({internal_name}) has wrong interface name in metadata.yaml "
                    f"(expected {self._expected_interfaces[internal_name]}, got {ep_meta.interface_name})"
                )

    def reconcile(self, ca_cert: Optional[str]):
        """Unconditional logic related to self-monitoring to run regardless of the event we are processing."""
        if self._charm_tracing.is_ready():
            endpoint = self._charm_tracing.get_endpoint("otlp_http")
            if not endpoint:
                return
            ops_tracing.set_destination(
                url=endpoint + "/v1/traces",
                ca=ca_cert,
            )
